import typing
# from hl7Lib import Hl7MessageValidator
# from .hl7MessageValidator import Hl7MessageValidator
import hl7

class Hl7MessageParser:
    """
    Parses hl7Messages and stores the extracted values in a dictionary

    By default, this parser does not extract any field, you must tell him what he needs to find and where it needs to find them.
    parser = Hl7MessageParser()
    parser.set_field_definition('PatientID', 'PID.F3.R1.C1')


    """

    def __init__(self):
        self._hl7_message = None

        self.fields_definitions = {
            '_encoding': 'MSH.F18'
        }

    def set_field_definition(self, field_name: str, key: str):
        """
        :param fieldName: the name of the field to extract.  You should use the Dicom Tag name (i.e.: 'AccessionNumber')
        :param key: the HL7 location of the field to extract (i.e: 'OBR.F18' or 'PID.F3.R1.C4')
        :return:
        """
        self.fields_definitions[field_name] = key

    def add_fields_definitions(self, definitions: typing.Dict[str, str]):
        """
        :param definitions: a dictionary like: {'PatientID': 'PID.F3.R1.C1'}
        """
        for key, value in definitions.items():
            self.fields_definitions[key] = value

    def parse(self, message: str, strict: bool = True) -> typing.Dict[str, str]:
        # message = Hl7MessageValidator().validate(message, strict)
        self._hl7_message = hl7.parse(message)

        values = {}

        for field_name, key in self.fields_definitions.items():
            value = self._extract_field(key)
            if value is not None:
                values[field_name] = value

        return values

    def _extract_field(self,
                      key: str,
                      segment_index: int = 0,
                      default_value: str = None
                      ) -> str:
        keys = key.split('.')
        if len(keys) == 2:
            return self._get_whole_field(key, segment_index, default_value = default_value)
        elif len(keys) == 4:
            assert (segment_index == 0)  # _get does not support segmentIndex yet
            return self._get(key, default_value = default_value)


    def _get(self, key: str, default_value: str = None) -> str:
        try:
            return self._hl7_message[key]
        except KeyError:
            return None
        except IndexError:
            return default_value

    def _get_whole_field(self, key: str, segment_index: int = 0, default_value: str = None):
        try:
            # key is something like 'OBR.F16'
            keys = key.split('.')
            assert (len(keys) == 2)
            assert (keys[1][0] == 'F')

            value = str(self._hl7_message[keys[0]][segment_index][int(keys[1][1:])])
            if value == '':
                return None
            return value
        except KeyError:
            return None
        except IndexError:
            return default_value



import subprocess
import time

from orthanc_api_client import OrthancApiClient, helpers
import pathlib
import logging
import unittest
import json
import requests

from orthanc_tools import LabelModifier

here = pathlib.Path(__file__).parent.resolve()

logger = logging.getLogger('orthanc_tools')
logger.setLevel(logging.DEBUG)
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
ch.setFormatter(formatter)
logger.addHandler(ch)


class TestLabelModifier(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        subprocess.run(["docker", "compose", "down", "-v"], cwd=here/"docker-setup-auth")
        subprocess.run(["docker", "compose", "up", "-d"], cwd=here/"docker-setup-auth")

        cls.oa = OrthancApiClient('http://localhost:10042', user='test', pwd='test')
        cls.oa.wait_started()

        cls.auth_url = "http://localhost:18000"
        cls.auth_user = "test"
        cls.auth_pwd = "test"

        # waiting for the auth service to be ready to accept queries
        cls.wait_auth_service_started(url=cls.auth_url, auth=(cls.auth_user, cls.auth_pwd))

        cls.permissions_old = {
          "roles": {
            "admin-role": {
              "authorized-labels": [
                "TEST",
                "OLD2"
              ],
              "permissions": [
                "all",
                "admin-permissions"
              ]
            },
            "doctor-role": {
              "authorized-labels": [
                "OLD",
                "TEST",
                "OLD2"
              ],
              "permissions": [
                "view",
                "download",
                "share",
                "send"
              ]
            },
            "external-role": {
              "authorized-labels": [
                "OLD"
              ],
              "permissions": [
                "view",
                "download"
              ]
            }
          },
          "available-labels": [
            "TEST",
            "OLD2",
            "OLD"
          ]
        }

        cls.permissions_new = {
          "roles": {
            "admin-role": {
              "authorized-labels": [
                "TEST",
                "OLD2"
              ],
              "permissions": [
                "all",
                "admin-permissions"
              ]
            },
            "doctor-role": {
              "authorized-labels": [
                "NEW",
                "TEST",
                "OLD2"
              ],
              "permissions": [
                "view",
                "download",
                "share",
                "send"
              ]
            },
            "external-role": {
              "authorized-labels": [
                "NEW"
              ],
              "permissions": [
                "view",
                "download"
              ]
            }
          },
          "available-labels": [
            "TEST",
            "OLD2",
            "NEW"
          ]
        }

        cls.permissions_new2 = {
          "roles": {
            "admin-role": {
              "authorized-labels": [
                "TEST",
                "OLD2"
              ],
              "permissions": [
                "all",
                "admin-permissions"
              ]
            },
            "doctor-role": {
              "authorized-labels": [
                "TEST",
                "OLD2"
              ],
              "permissions": [
                "view",
                "download",
                "share",
                "send"
              ]
            },
            "external-role": {
              "authorized-labels": [
                "OLD2"
              ],
              "permissions": [
                "view",
                "download"
              ]
            }
          },
          "available-labels": [
            "TEST",
            "OLD2"
          ]
        }

    @classmethod
    def tearDownClass(cls):
        subprocess.run(["docker", "compose", "down", "-v"], cwd=here/"docker-setup-auth")

    @classmethod
    def wait_auth_service_started(cls, url, auth):
        ready = False
        while not ready:
            try:
                response = requests.get(url=url + "/settings/roles", auth=auth)
                if response.status_code == 200:
                    ready = True
            except Exception as ex:
                time.sleep(1)

    def sorting(self, item):
        if isinstance(item, dict):
            return sorted((key, self.sorting(values)) for key, values in item.items())
        if isinstance(item, list):
            return sorted(self.sorting(x) for x in item)
        else:
            return item

    def test_apply_new_label(self):

        # clean up
        self.oa.delete_all_content()

        # upload an instance
        self.oa.upload_file(here / "stimuli/CT_small.dcm")

        old_label = "OLD"
        new_label = "NEW"

        # apply label on this study
        study_id = self.oa.studies.get_all_ids()[0]
        self.oa.studies.add_label(study_id, old_label)

        modifier = LabelModifier(api_client=self.oa, auth_service_url=self.auth_url, auth_service_login=self.auth_user, auth_service_password=self.auth_pwd)

        modifier.apply_new_label(new_label=new_label, old_label=old_label)

        self.assertEqual(len(self.oa.studies.find(query={'PatientID': '*'}, labels=[new_label])), 1)
        self.assertEqual(len(self.oa.studies.find(query={'PatientID': '*'}, labels=[old_label])), 0)


    def test_modify_permissions(self):

        # reset permissions.json file to original values
        with open(here / "docker-setup-auth/permissions.json", 'w') as f:
            f.write(json.dumps(self.permissions_old))

        subprocess.run(["docker", "compose", "restart", "orthanc-auth-service"], cwd=here / "docker-setup-auth")

        # waiting for the auth service to be ready to accept queries
        self.wait_auth_service_started(self.auth_url, (self.auth_user, self.auth_pwd))

        modifier = LabelModifier(api_client=self.oa, auth_service_url=self.auth_url, auth_service_login=self.auth_user, auth_service_password=self.auth_pwd)

        old_label = "OLD"
        new_label = "NEW"

        modifier.modify_permissions(old_label=old_label, new_label=new_label)
        modifier.add_label_to_available_list(label_to_add=new_label)
        modifier.remove_label_from_available_list(label_to_remove=old_label)

        current_permissions = modifier.get_roles()
        self.assertEqual(self.sorting(current_permissions), self.sorting(self.permissions_new))

    def test_modify_permissions_with_existing_label(self):
        '''
        To check that a label is not twice in a list
        '''

        # reset permissions.json file to original values
        with open(here / "docker-setup-auth/permissions.json", 'w') as f:
            f.write(json.dumps(self.permissions_old))

        # restart auth service to take file in account
        subprocess.run(["docker", "compose", "restart", "orthanc-auth-service"], cwd=here / "docker-setup-auth")

        # waiting for the auth service to be ready to accept queries
        self.wait_auth_service_started(self.auth_url, (self.auth_user, self.auth_pwd))

        modifier = LabelModifier(api_client=self.oa, auth_service_url=self.auth_url, auth_service_login=self.auth_user, auth_service_password=self.auth_pwd)

        old_label = "OLD"
        new_label = "OLD2"

        modifier.modify_permissions(old_label=old_label, new_label=new_label)
        modifier.add_label_to_available_list(label_to_add=new_label)
        modifier.remove_label_from_available_list(label_to_remove=old_label)

        current_permissions = modifier.get_roles()
        self.assertEqual(self.sorting(current_permissions), self.sorting(self.permissions_new2))

if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    unittest.main()




import queue
import threading
import time
import os
import logging

from orthanc_api_client import OrthancApiClient, ChangeType
from .helpers.scheduler import Scheduler
logger = logging.getLogger(__name__)


class OrthancMonitor:
    """
    Monitor the /changes route and trigger callback when a new change is detected
    """

    def __init__(self, 
                 api_client: OrthancApiClient,
                 worker_threads_count: int = 1,
                 persist_status_path: str = None,
                 start_at_sequence_id: int = None,
                 polling_interval: float = 0.5,
                 scheduler: Scheduler = None,
                 max_retries: int = 5,
                 error_folder_path: str = None
    ):

        self._api_client = api_client
        self._changes_to_process = queue.Queue(worker_threads_count + 1)
        self._monitoring_thread = None
        self._worker_threads_count = worker_threads_count
        self._worker_threads = []
        self._is_running = False
        self._polling_interval = polling_interval
        self._handlers = {}
        self._persist_status_path = None
        self._persist_status_lock = threading.RLock()
        self._changes_id_being_processed = set()
        self._largest_processed_change_id = 0
        self._largest_ignored_change_id = 0
        self._scheduler = scheduler
        self._error_folder_path = error_folder_path
        self._max_retries = max_retries

        if persist_status_path is not None:
            self._persist_status_path = persist_status_path
            self._start_at_sequence_id = self._read_status_from_file()

        elif start_at_sequence_id is not None:
            self._start_at_sequence_id = start_at_sequence_id
        else:
            self._start_at_sequence_id = 0
        

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    def add_handler(self, change_type: ChangeType, callback):
        self._handlers[change_type] = callback

    def _read_status_from_file(self):
        try:
            with open(self._persist_status_path) as f:
                sequence_id = int(f.read())
        except (ValueError, FileNotFoundError):  # if can not read, start at 0
            logger.warning("Could not read sequence id from file, starting at 0")
            return 0

        logger.info(f"Starting at sequence id from file = {sequence_id}")
        return sequence_id

    def _mark_change_as_being_processed(self, sequence_id):
        if self._persist_status_path is None:
            return

        with self._persist_status_lock:
            logger.debug(f"marking change {sequence_id} as being processed")
            self._changes_id_being_processed.add(sequence_id)

    def _mark_change_as_processed(self, sequence_id, has_ignored_change):
        # note: this can be improved: if multiple workers are processing events, we might skip a few changes when restarting
        if self._persist_status_path is None:
            return
        
        with self._persist_status_lock:
            self._changes_id_being_processed.remove(sequence_id)

            self._largest_processed_change_id = max(self._largest_processed_change_id, sequence_id)
            if len(self._changes_id_being_processed) > 0:
                restart_at_sequence_id = min(self._changes_id_being_processed) - 1
                if not has_ignored_change:
                    logger.debug(f"marking change {sequence_id} as processed, will restart at {restart_at_sequence_id}, changes being processed: " + ", ".join([str(c) for c in self._changes_id_being_processed]))
            else:
                restart_at_sequence_id = max(self._largest_processed_change_id, self._largest_ignored_change_id)
                if not has_ignored_change:
                    logger.debug(f"marking change {sequence_id} as processed, will restart at {restart_at_sequence_id}")

            # first write to a temp file and then move the file to make the operation robust
            tmp = self._persist_status_path + ".tmp"
            try:
                with open(tmp, "wt") as f:
                    f.write(str(restart_at_sequence_id))
                os.replace(tmp, self._persist_status_path)  # this is an 'atomic' operation
            except OSError as ex:
                raise Exception(f"Could not write sequence id to file \"{ex.filename}\": {ex.strerror}")

    def start(self, existing_changes_only: bool=False):
        """
        Parameters:
            existing_changes_only: True to stop processing changes once all current changes have been processed
                                   False to continue monitoring for new changes
        """
        # create monitoring thread
        self._monitoring_thread = threading.Thread(
            target=self._monitor_changes,
            name='Monitoring Thread',
            args=(existing_changes_only, )
        )

        # create worker threads
        for thread_id in range(0, self._worker_threads_count):
            self._worker_threads.append(threading.Thread(
                target=self._process_changes,
                name=f"Worker Thread {thread_id}",
                args=(thread_id, )
            ))

        # start threads
        self._is_running = True
        self._monitoring_thread.start()
        for wt in self._worker_threads:
            wt.start()

    def stop(self):
        logger.debug("Stopping Orthanc Monitor")

        # first stop the monitoring thread so we don't produce events anymore
        self._is_running = False
        self._monitoring_thread.join()

        # post one 'empty' exit message per thread to unlock the threads from waiting on the process queue
        for i in range(0, self._worker_threads_count):
            self._changes_to_process.put(None)

        for t in self._worker_threads:
            t.join()

    def _monitor_changes(self, existing_changes_only):
        logger.debug(f"Starting Monitoring thread at change id = {self._start_at_sequence_id}")

        last_sequence_id = self._start_at_sequence_id
        done = False
        has_logged_connection_error = False  # to avoid repeating errors

        while self._is_running:
            if existing_changes_only and done:
                self._is_running = False
                return

            done = False

            while not done and self._is_running:  # read as fast as you can while there are still events

                if self._scheduler:
                    self._scheduler.wait_right_time_to_run()

                # get the list of changes from orthanc
                try:
                    changes, last_sequence_id, done = self._api_client.get_changes(
                        since=last_sequence_id,
                        limit=100
                    )
                    if has_logged_connection_error:
                        has_logged_connection_error = False
                        logger.warning("Connected to Orthanc again")
                except Exception as ex:
                    if not has_logged_connection_error:
                        logger.warning("Could not reach Orthanc, retrying silently...")
                        has_logged_connection_error = True
                    break

                # enqueue the events
                for change in changes:
                    if change.change_type in self._handlers: # enqueue only the changes that will be handled
                        self._mark_change_as_being_processed(change.sequence_id)
                        self._changes_to_process.put(change)  # if the queue is full, this will block until there's a free slot
                    else:
                        self._largest_ignored_change_id = change.sequence_id

            # if no events available, wait and poll again
            time.sleep(self._polling_interval)

    def _process_changes(self, worker_id):
        logger.debug(f"Starting Processing thread {worker_id}")

        while True:
            change = self._changes_to_process.get()  # block until a message is available

            if change is None:  # sent by stop() to stop all worker threads
                self._changes_to_process.task_done()
                break

            retries = 0
            retry_delays = [5, 20, 60, 300, 900, 1800, 3600, 7200]
            processed = False
            last_error = None

            while not processed and retries <= min(self._max_retries, len(retry_delays)):
                if retries >= 1:
                    delay = retry_delays[retries - 1]
                    logger.info(f"waiting {delay} seconds before retrying change {change.sequence_id} {change.change_type}")
                    time.sleep(delay)

                try:
                    # process events (this is blocking the worker thread until the handler returns)
                    if change.change_type in self._handlers:
                        logger.info(f"processing change {change.sequence_id} {change.change_type}")
                        self._handlers[change.change_type](change.sequence_id, change.resource_id, self._api_client)
                        processed = True

                except Exception as ex:
                    logger.exception("Unhandled exception in event handler: ", ex)
                    last_error = str(ex)

                retries = retries + 1


            if not processed and self._error_folder_path:
                error_file_path = os.path.join(self._error_folder_path, f"{change.sequence_id:010d}." + str(change.change_type) + ".error.txt")
                try:
                    with open(error_file_path, "wt") as f:
                        f.write(last_error)
                except OSError as ex:
                    raise Exception(f"Could not write error report to file \"{ex.filename}\": {ex.strerror}")
                # if we store errors on disk, we consider that the change has been processed since we keep a track of its failure -> it will not be handled again after a restart
                processed = True

            if processed:
                self._mark_change_as_processed(change.sequence_id, has_ignored_change=False)

            self._changes_to_process.task_done()  # tell the queue the item has been processed

        logger.debug("Processing thread stopped")


    def execute(self, existing_changes_only: bool = True):
        """
        Parameters:
            existing_changes_only: True to stop processing changes once all current changes have been processed
                                   False to continue monitoring for new changes
        """

        self.start(existing_changes_only=existing_changes_only)

        if existing_changes_only:
            while self._is_running:
                time.sleep(self._polling_interval)
            self.stop()



import time


class Timer:
    """
    Helper to measure time duration (in seconds/ms)

    example:
        timer = Timer()  # creates and starts the Timer
        lengthyOperation()
        elapsed = timer.get_elapsed_seconds() # get the elapsed time.  The timer continues to measure the elapsed time
        lengthyOperation()
        elapsed = timer.get_elapsed_seconds() # get the elapsed time since beginning

        timer.reset()  # reset the timer
    """


    def __init__(self):
        self.reset()

    def reset(self):
        self._start = time.perf_counter()

    def get_elapsed_seconds(self):
        stop = time.perf_counter()
        return stop - self._start

    def get_elapsed_ms(self):
        return 1000.0 * self.get_elapsed_seconds()



import contextlib
import typing
# from hl7Lib import Hl7MessageParser
from .hl7_message_parser import Hl7MessageParser


class Hl7WorklistParser(Hl7MessageParser):

    def __init__(self, specific_fields: dict = None, patient_name_components_count: int = 5):
        super(Hl7WorklistParser, self).__init__()

        self._patient_name_components_count = patient_name_components_count

        # Let's add "standards" field
        self.add_fields_definitions({
            # --- PID segment
            'PatientID': 'PID.F3.R1.C1',
            'IssuerOfPatientID': 'PID.F3.R1.C4',
            'OtherPatientIDs': 'PID.F4.R1.C1',
            'PatientName': 'PID.F5',  # IHE recommend to use the whole Field 5 as the patient name (it's usually full of ^ that is a separator in HL7 so we must take it as is without trying to parse it !
            'PatientMotherBirthName': 'PID.F6',
            'PatientBirthDate': 'PID.F7',
            '_sex': 'PID.F8',
            'PatientSpeciesDescription': 'PID.F9',
            'PatientBreedDescription': 'PID.F10',
            'PatientAddress': 'PID.F11',
            'PatientSexNeutered': 'PID.F27', # from the theory, PID.F27 is 'Veterans Military Status', we decided to use it with AssistoVet

            # --- OBR segment
            #'AccessionNumber': 'OBR.F18',
            'AccessionNumber': 'OBR.F3.R1.C1',
            #'PatientState': 'OBR.F12', # Seems to be something describing the state of the patient during the examination, so we don't put it in the worklist
            '_requestingPhysicianOBR': 'OBR.F16',
            'ReasonForTheRequestedProcedure': 'OBR.F31',
            'Modality': 'OBR.F24',
            'RequestedProcedureDescription': 'OBR.F4.R1.C2',
            #'ScheduledProcedureStepStartDateTime': 'OBR.F27.R1.C4',
            '_scheduledProcedureStepStartDateTime': 'OBR.F27.R1.C4',
            '__scheduledProcedureStepStartDateTime': 'OBR.F20', # By default, this is the segment used by assistovet it seems not to be in use in other cases

            # --- PV1 segment
            '_ambulatoryStatus': 'PV1.F15',
            'ReferringPhysicianName': 'PV1.F8',
            'ConfidentialityConstraintOnPatientDataDescription': 'PV1.F16',
            '_consultingDoctorPV1': 'PV1.F9',
            ''

            # --- ORC segment
            'OrderPlacerIdentifierSequence': 'ORC.F2.R1.C1', # not used in any builder!
            'RequestedProcedureID': 'ORC.F2.R1.C1',
            'ScheduledProcedureStepID': 'ORC.F2.R1.C1',
            'OrderFillerIdentifierSequence': 'ORC.F3', # not used in any builder!
            '_requestingPhysicianORC': 'ORC.F12',

            # --- ZDS segment
            'StudyInstanceUID': 'ZDS.F1.R1.C1'
        })

        # Let's add specific fields (they will override the default ones)
        if specific_fields is not None:
            self.add_fields_definitions(specific_fields)

    def parse(self, hl7_message: str) -> typing.Dict:

        # set a bunch of default values to make sure worklists are accepted by some GE modalities
        # REQUIRED elements
        values = {}
        values['AccessionNumber'] = None
        values['RequestingPhysician'] = None
        values['RequestedProcedureDescription'] = None
        values['Modality'] = None
        values['ReferringPhysicianName'] = None
        values['ScheduledStationAETitle'] = 'UNKNOWN'
        values['ScheduledPerformingPhysicianName'] = None
        values['ScheduledStationName'] = None
        values['ScheduledProcedureStepID'] = 'UNKNOWN'
        values['RequestedProcedureID'] = 'UNKNOWN'
        values['SpecificCharacterSet'] = 'ASCII'

        # extract field the default way
        values_from_hl7 = super(Hl7WorklistParser, self).parse(hl7_message, strict = False)

        values.update(values_from_hl7)

        # keep only the first 5 components of the name according to http://dicom.nema.org/dicom/2013/output/chtml/part05/sect_6.2.html (check PN VR definition)
        values['PatientName'] = '^'.join(values['PatientName'].split('^')[:self._patient_name_components_count])

        sex = values['_sex']
        if sex is None or sex in ['M', 'F']:
            values['PatientSex'] = sex
        elif sex in ['U']:  # unknown in HL7 -> empty in DICOM
            values['PatientSex'] = ''
        elif sex in ['A', 'N', 'O']:  # Ambiguous or Not Applicable or Other in HL7 -> 'other' in Dicom
            values['PatientSex'] = 'O'

        # clean birthdate
        values['PatientBirthDate'] = values['PatientBirthDate'][0:8]

        if values.get('_ambulatoryStatus') is not None and 'B6' in values['_ambulatoryStatus']:
            values['PregnancyStatus'] = 3

        # --- OBX segment parsing
        with contextlib.suppress(KeyError):  # OBX segments might not be there
            for i in range(0, len(self._hl7_message['OBX'])):
                observation = self._get_whole_field('OBX.F3', segment_index = i)
                if 'BODY WEIGHT' in observation.upper():
                    values['PatientWeight'] = self._get_whole_field('OBX.F5', segment_index = i)
                if 'BODY HEIGHT' in observation.upper():
                    values['PatientHeight'] = self._get_whole_field('OBX.F5', segment_index = i)

        if values.get('_encoding') in [None, '8859/1', '8859/15']:
            values['SpecificCharacterSet'] = 'ISO_IR 100'

        if values.get('_scheduledProcedureStepStartDateTime') is not None:
            datetimeString = values.get('_scheduledProcedureStepStartDateTime')
            values['ScheduledProcedureStepStartDate'] = datetimeString[:8]  # date is made of the 8 first chars of the string
            if len(datetimeString) == 12:
                values['ScheduledProcedureStepStartTime'] = datetimeString[8:12] + "00"
            elif len(datetimeString) == 14:
                values['ScheduledProcedureStepStartTime'] = datetimeString[8:14]
        elif values.get('__scheduledProcedureStepStartDateTime') is not None: # AssistoVet case
            datetimeString = values.get('__scheduledProcedureStepStartDateTime')
            values['ScheduledProcedureStepStartDate'] = datetimeString[
                                                        :8]  # date is made of the 8 first chars of the string
            values['ScheduledProcedureStepStartTime'] = datetimeString[8:14]



        if values.get('_requestingPhysicianOBR') is not None:
            values['RequestingPhysician'] = values.get('_requestingPhysicianOBR')
        elif values.get('_requestingPhysicianORC') is not None:
            values['RequestingPhysician'] = values.get('_requestingPhysicianORC')
        elif values.get('_consultingDoctorPV1') is not None:
            values['RequestingPhysician'] = values.get('_consultingDoctorPV1')

        # specific to AssistoVet, this is the first Vet HL7 messages provider integrated, so this could elvove in the future:
        if values.get('PatientSpeciesDescription') is not None and values.get('PatientBreedDescription') is not None:
            # this should be an animal, so let's fill the responsible name
            last_name = self._get('PID.F5.R1.C1', default_value="")
            first_name = values.get('PatientMotherBirthName')
            if first_name is None: first_name = ""
            values['ResponsiblePerson'] = '^'.join([last_name, first_name])

        return values




import os


class Hl7Error(Exception):
    def __init__(self, message, hl7_request = None, hl7_response = None):
        self.message = message
        self.hl7_request = hl7_request
        self.hl7_response = hl7_response

    def __str__(self):
        return os.linesep.join(['HL7 client exception: {0}'.format(self.message),
                                'request = {0}'.format(str(self.hl7_request).replace('\r', os.linesep)),
                                'response = {0}'.format(str(self.hl7_response).replace('\r', os.linesep))])


class UnsupportedMessageType(Exception):
    """
    Error that occurs when the :class:`MLLPServer` receives a message without an associated handler
    """

    def __init__(self, msgType):
        self._msgType = msgType

    def __str__(self):
        return 'No Handler found for message type %s' % self._msgType


class InvalidHL7Message(Exception):
    """
    Error that occurs when the :class:`MLLPServer` receives a string which doesn't represent an ER7-encoded HL7 message
    """

    def __init__(self, msg):
        self._msg = msg

    def __str__(self):
        return 'The string received is not a valid HL7 message : {0}'.format(self._msg)



import time, sys
import argparse
import logging
from orthanc_api_client import OrthancApiClient
import os

import schedule

# examples:
# python orthanc_tools/orthanc_warmer.py --url=http://192.168.0.10:8042 --user=user --password=pwd --interval=30


logger = logging.getLogger(__name__)

class OrthancWarmer:
    def __init__(self,
                 api_client: OrthancApiClient,
                 interval: int
                 ):
        self._api_client = api_client
        self._interval = interval
        self._errors_counter = 0

    def find(self):
        try:
            o.studies.find(query={'StudyDate': "19500101"})
            # 1st of January 1950 is sunday, so probably no results
            self._errors_counter = 0
            logger.debug("Find succeeded!")

        except Exception as ex:
            self._errors_counter += 1
            logger.error(f"{str(ex)}")
            if self._errors_counter > 5:
                logger.error("Last 5 queries to Orthanc failed, so exiting...")
                sys.exit()


    def execute(self):
        logger.info("----- Initializing Orthanc Warmer...")
        schedule.every(interval).minutes.do(self.find)

        while True:
            schedule.run_pending()
            time.sleep(1)

if __name__ == '__main__':
    level = logging.INFO
    if os.environ.get('VERBOSE_ENABLED'):
        level = logging.DEBUG
    logging.basicConfig(level=level, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    parser = argparse.ArgumentParser(description='Periodically queries Orthanc to keep the database warm.')
    parser.add_argument('--url', type=str, default='http://localhost:8042', help='Orthanc url')
    parser.add_argument('--user', type=str, default=None, help='Orthanc user name')
    parser.add_argument('--password', type=str, default=None, help='Orthanc password')
    parser.add_argument('--api_key', type=str, default=None, help='Orthanc api-key')
    parser.add_argument('--interval', type=int, default=30, help='Period of time between 2 queries (minutes).')
    args = parser.parse_args()

    url = os.environ.get("ORTHANC_URL", args.url)
    user = os.environ.get("ORTHANC_USER", args.user)
    password = os.environ.get("ORTHANC_PWD", args.password)
    api_key = os.environ.get("ORTHANC_API_KEY", args.api_key)
    interval = int(os.environ.get("INTERVAL", args.interval))

    o = None
    if api_key is not None:
        o=OrthancApiClient(url, headers={"api-key":api_key})
    else:
        o=OrthancApiClient(url, user=user, pwd=password)

    warmer = OrthancWarmer(api_client=o, interval=interval)
    warmer.execute()


from .hl7_error import Hl7Error, UnsupportedMessageType, InvalidHL7Message
from .hl7 import HL7_FIELD_SEPARATOR, HL7_ALL_SEPARATORS
from .hl7_message_validator import Hl7MessageValidator
from .hl7_message_parser import Hl7MessageParser
from .hl7_worklist_parser import Hl7WorklistParser
from .hl7_dicom_worklist_builder import DicomWorklistBuilder
from .hl7_client import MLLPClient
from .hl7_server import MLLPServer
from .hl7_orm_worklist_msg_handler import Hl7OrmWorklistMsgHandler
from .hl7_oru_report_msg_handler import Hl7OruReportMsgHandler
from .hl7_report_series_builder import ReportSeriesBuilder
from .hl7_report_parser import Hl7ReportParser



import argparse
import logging
from orthanc_api_client import OrthancApiClient
from typing import List
import zipfile
import tempfile
import os, time, sys
import multiprocessing
import queue
import threading

# examples:
# python orthanc_tools/orthanc_folder_importer.py --folder=./tests/stimuli --url=http://192.168.0.10:8042 --user=user --password=pwd --skip=.txt,.ini


logger = logging.getLogger(__name__)

class OrthancFolderImporter:
    '''
    Upload all the DICOM files contained in a folder (and its sub folders).
    It is a little bit smart:
    - There is a retry for every file and when it fails anyway, the file path is logged, but the script keeps working
    - Every sub folder uploaded (even with errors for some files) is logged, so that if the script is interrupted and
    restarted, it will restart from the last succeeded folder.
    - Zip files are unziped before upload
    '''
    def __init__(self,
                 api_client: OrthancApiClient,
                 folder_path: str,
                 errors_path: str,
                 state_path: str,
                 labels_list: List[str] = None,
                 max_retries: int = 8,
                 worker_threads_count: int = multiprocessing.cpu_count() - 1  # by default, use all CPUs but one for compression
                 ):
        self._api_client = api_client
        self._folder_path = folder_path
        self._labels_list = labels_list
        self._errors_path = errors_path # will contain the list of all the files path not correctly uploaded
        self._state_path = state_path # will contain the list of all the folders correctly uploaded

        self._worker_threads_count = worker_threads_count
        self._worker_threads = []
        self._messages = queue.Queue(maxsize=2*worker_threads_count)  # this is thread safe https://docs.python.org/3.5/library/queue.html#module-queue

        self._folders_uploaded = []

        if max_retries > 8:
            self._max_retries = 8
        else:
            self._max_retries = max_retries

        self._lock = threading.Lock()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    def add_file_name_in_errors_log(self, file_path):
        with open(self._errors_path, "at") as f:
            f.write(file_path + "\n")

    def add_folder_path_in_state_file(self, folder_path):
        with self._lock:
            with open(self._state_path, "at") as f:
                f.write(folder_path + "\n")

    def upload_and_label(self, path_to_upload):
        """
        Upload the file if path_to_upload is a file path
        Recursively upload the content of the folder is path_to_upload is a folder path
        Then apply the labels on the study
        """

        # file path case
        if os.path.isfile(path_to_upload):

            # zip file case
            if "zip" in path_to_upload and zipfile.is_zipfile(path_to_upload):
                with tempfile.TemporaryDirectory() as tempDir:
                    with zipfile.ZipFile(path_to_upload, 'r') as z:
                        z.extractall(tempDir)
                    for path in os.listdir(tempDir):
                        full_path = os.path.join(tempDir, path)
                        self.upload_and_label(path_to_upload=full_path)
            else:
                retry_count = 0
                retry_delays = [5, 20, 60, 300, 900, 1800, 3600, 7200]

                while retry_count <= self._max_retries:
                    if retry_count >= 1:
                        delay = retry_delays[retry_count - 1]
                        logger.info(f"waiting {delay} seconds before retrying the upload of {path_to_upload}")
                        time.sleep(delay)
                    try:
                        # here, we should have only files (and no zip file)
                        instance_orthanc_ids = self._api_client.upload_file(path_to_upload, ignore_errors=True)
                        if len(instance_orthanc_ids) == 0:
                            logger.error(f"File not uploaded: {path_to_upload}.")
                            self.add_file_name_in_errors_log(file_path=path_to_upload)
                            break
                        # we label for each instance, not at the end of the study, so that there is never an unlabeled image in Orthanc
                        if self._labels_list is not None:
                            study_orthanc_id = self._api_client.instances.get_parent_study_id(instance_orthanc_ids[0])
                            self._api_client.studies.add_labels(orthanc_id=study_orthanc_id, labels=self._labels_list)
                        break
                    except Exception as e:
                        if retry_count == self._max_retries:
                            logger.error(f"Error while uploading this file: {path_to_upload}. Exception: {str(e)}")
                            logger.error(f"too many attempts, logging the file name...")
                            self.add_file_name_in_errors_log(file_path=path_to_upload)
                            break
                        else:
                            retry_count += 1
                            logger.warning(f"Error while uploading this file, retrying...: {path_to_upload}. Exception: {str(e)}")
        # folder case
        elif os.path.isdir(path_to_upload):
            # this folder could have been processed in a previous run of the script
            if path_to_upload in self._folders_uploaded:
                logger.info(f"Folder {path_to_upload} already processed, skipping...")
                return

            # let's process this folder
            for path in os.listdir(path_to_upload):
                full_path = os.path.join(path_to_upload, path)
                self.upload_and_label(path_to_upload=full_path)

            # let's add this folder path in the processed ones:
            self.add_folder_path_in_state_file(path_to_upload)

    def _process_path(self, worker_id):
        logger.debug(f"Starting Processing thread {worker_id}")

        while True:
            path = self._messages.get()  # block until a message is available

            if path is None:  # sent by stop() to stop all worker threads
                self._messages.task_done()
                break

            # path is the full path of a file or a folder
            self.upload_and_label(path_to_upload=path)

            self._messages.task_done()  # tell the queue the item has been processed

        logger.debug("Processing thread stopped")

    def execute(self):
        # read state
        if os.path.isfile(self._state_path):
            with open(self._state_path, 'r') as file:
                lines = file.readlines()
                self._folders_uploaded = [line.strip() for line in lines]

        # create worker threads
        for thread_id in range(0, self._worker_threads_count):
            self._worker_threads.append(threading.Thread(
                target=self._process_path,
                name=f"Worker Thread {thread_id}",
                args=(thread_id, )
            ))

        # start threads
        for wt in self._worker_threads:
            wt.start()

        # let's browse the main folder to feed the message queue
        for path in os.listdir(path=self._folder_path):
            full_path = os.path.join(self._folder_path, path)
            self._messages.put(full_path) # if the queue is full, this will block until there's a free slot

        # let's wait for the completion of all threads
        self.stop()

        logger.info("End of upload!")

    def stop(self):
        logger.info("Waiting for Orthanc Folder Importer to complete each upload...")

        # post one 'empty' exit message per thread to unlock the threads from waiting on the process queue
        for i in range(0, self._worker_threads_count):
            self._messages.put(None)

        for t in self._worker_threads:
            t.join()

if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    parser = argparse.ArgumentParser(description='Import the content of a folder in Orthanc')
    parser.add_argument('--url', type=str, default='http://localhost:8042', help='Orthanc url')
    parser.add_argument('--user', type=str, default=None, help='Orthanc user name')
    parser.add_argument('--password', type=str, default=None, help='Orthanc password')
    parser.add_argument('--api_key', type=str, default=None, help='Orthanc api-key')
    parser.add_argument('--folder_path', type=str, help='Folder to import, the one containing the DICOM files.')
    parser.add_argument('--labels_list', type=str, default=None, help='List of labels to apply to the uploaded studies, separated by a comma.')
    parser.add_argument('--errors_path', type=str, help='Path of the file which will contain the list of problematic files (not uploaded).')
    parser.add_argument('--state_path', type=str, help='Path of the file which will contain the list of all the folder correctly uploaded.')
    parser.add_argument('--max_retries', type=int, default=8, help='Maximum number of attempts for a file upload.')
    parser.add_argument('--worker_threads_count', type=int, default=1, help='Worker threads count')

    args = parser.parse_args()

    url = os.environ.get("ORTHANC_URL", args.url)
    user = os.environ.get("ORTHANC_USER", args.user)
    password = os.environ.get("ORTHANC_PWD", args.password)
    api_key = os.environ.get("ORTHANC_API_KEY", args.api_key)
    folder_path = os.environ.get("FOLDER_PATH", args.folder_path)
    labels_list = os.environ.get("LABELS_LIST", args.labels_list)
    errors_path = os.environ.get("ERRORS_PATH", args.errors_path)
    state_path = os.environ.get("STATE_PATH", args.state_path)
    max_retries = int(os.environ.get("MAX_RETRIES", str(args.max_retries)))
    worker_threads_count = int(os.environ.get("WORKER_THREADS_COUNT", str(args.worker_threads_count)))

    o = None
    if api_key is not None:
        o=OrthancApiClient(url, headers={"api-key": api_key})
    else:
        o=OrthancApiClient(url, user=user, pwd=password)

    importer = OrthancFolderImporter(
        api_client=o,
        folder_path=folder_path,
        labels_list=labels_list,
        errors_path=errors_path,
        state_path=state_path,
        max_retries=max_retries,
        worker_threads_count=worker_threads_count
    )

    importer.execute()



import json
import pprint

import inquirer
import logging
import argparse
from orthanc_api_client import OrthancApiClient
import sys
import requests
logger = logging.getLogger(__name__)

'''
WARNING
-------
This script is interactive!
So, run it from the console and follow instructions...

PURPOSE
-------
This script allows to handle the fix of a typo in a label name.

HOW IT WORKS
------------
    ask for the creds

    show the list of labels and ask for the one to change

    ask for the new label value

    add the new label to the list of available labels (if this list is not empty)

    apply new label to all the studies from existing one

    remove original label from all the studies from the new one

    handle permissions
    
    remove the old label from the list of available labels (if this list is not empty)

HOW TO USE IT
-------------
- get the url and creds for both Orthanc and Auth-service
- Run it thanks to:
`python3 label_modifier.py --orthanc_url=http://localhost/orthanc-api/ --orthanc_user=demo --auth_url=http://localhost/auth-service --auth_user=demo`
- Note that password will be prompted by the script itself
'''

class LabelModifier:
    def __init__(self,
                 api_client: OrthancApiClient,
                 auth_service_url: str,
                 auth_service_login: str,
                 auth_service_password: str
                 ):
        self._api_client = api_client
        self._auth_service_url = auth_service_url
        self._auth_service_login = auth_service_login
        self._auth_service_password = auth_service_password

        self._auth_service_roles_url = self._auth_service_url + "/settings/roles"
        self._auth_service_auth = (self._auth_service_login, self._auth_service_password)


    def apply_new_label(self, new_label: str, old_label: str):

        # Here is a loop because the find will be limited to LimitFindResults value
        while True:
            # find all the studies with the old label
            studies = self._api_client.studies.find(
                query={},
                labels=[old_label]
            )
            if len(studies) == 0:
                return

            # add the new label and then delete the old one
            for study in studies:
                self._api_client.studies.add_label(study.orthanc_id, new_label)
                self._api_client.studies.delete_label(study.orthanc_id, old_label)

    def get_roles(self):

        # get the roles/perm
        response = requests.get(self._auth_service_roles_url, auth=self._auth_service_auth)

        # Check if request was successful
        if response.status_code == 200:
            # get json (as a string)
            return response.json()
        else:
            logger.error("Error getting permissions:", response.status_code)
            sys.exit(1)

    def modify_permissions(self, new_label: str, old_label: str):

        content_as_json = self.get_roles()

        # replace old label by new one
        modified_content = self.update_dict_values(content_as_json, "authorized-labels", old_label, new_label)

        # push the modifications
        response = requests.put(
            url=self._auth_service_roles_url,
            json=modified_content,
            auth=self._auth_service_auth)

        if response.status_code != 200:
            logger.error("Error setting permissions:", response.status_code)
            sys.exit(1)

    def add_label_to_available_list(self, label_to_add: str):
        content_as_json = self.get_roles()

        # get available-labels
        available_labels = content_as_json["available-labels"]

        # if there is no restriction on labels, let's forget it
        if len(available_labels) == 0:
            return

        # let's add the new label to the list (without duplicate)
        temp_list = content_as_json["available-labels"]
        temp_list.append(label_to_add)
        content_as_json["available-labels"] = list(dict.fromkeys(temp_list))

        # push the modifications
        response = requests.put(
            url=self._auth_service_roles_url,
            json=content_as_json,
            auth=self._auth_service_auth)

        if response.status_code != 200:
            logger.error("Error setting permissions:", response.status_code)
            sys.exit(1)

    def remove_label_from_available_list(self, label_to_remove: str):
        content_as_json = self.get_roles()

        # get available-labels
        available_labels = content_as_json["available-labels"]

        # if there is no restriction on labels, let's forget it
        if len(available_labels) == 0:
            return

        # let's remove the old label from the list
        content_as_json["available-labels"].remove(label_to_remove)

        # push the modifications
        response = requests.put(
            url=self._auth_service_roles_url,
            json=content_as_json,
            auth=self._auth_service_auth)

        if response.status_code != 200:
            logger.error("Error setting permissions:", response.status_code)
            sys.exit(1)

    def update_dict_values(self, data, target_key, old_value, new_value):
        """
        Searches for a specific key in a nested dictionary, updates the target value in arrays if the key is found.

        :param data: The dictionary to search.
        :param target_key: The key to find in the dictionary.
        :param old_value: The value to replace in the arrays.
        :param new_value: The new value to use for replacement.
        :return: Updated dictionary.
        """
        if isinstance(data, dict):
            for key, value in data.items():
                if key == target_key and isinstance(value, list):
                    # Replace old_value with new_value in the array
                    updated_list = [new_value if item == old_value else item for item in value]
                    data[key] = list(dict.fromkeys(updated_list))  # Remove duplicates
                elif isinstance(value, (dict, list)):
                    # Recursively process nested dictionaries and lists
                    self.update_dict_values(value, target_key, old_value, new_value)
        elif isinstance(data, list):
            for item in data:
                if isinstance(item, (dict, list)):
                    # Recursively process lists of dictionaries or nested lists
                    self.update_dict_values(item, target_key, old_value, new_value)
        return data

    def execute(self):

        # let's get the existing labels
        labels_list = self._api_client.get_all_labels()
        label_to_modify = None

        # let the user choose the label to modify
        if len(labels_list)>0:
            questions_labels = [
                inquirer.List(
                    "orthanc_labels_chosen_list",
                    message="Choose the label to modify",
                    choices=labels_list,
                )
            ]

            answers = inquirer.prompt(questions_labels)
            label_to_modify = answers['orthanc_labels_chosen_list']
        else:
            logger.error(f"No labels found, exiting...")
            sys.exit(1)

        # let the user choose the new value for the label to modify
        questions_labels = [
            inquirer.Text(
                "new_label_value",
                message="Write the new value for the label",
            )
        ]

        answers = inquirer.prompt(questions_labels)
        label_new_value = answers['new_label_value']

        print("-----------------------------------------------------------------------------------")
        print(f"This change will be applied:\r\n   \"{label_to_modify}\" will become \"{label_new_value}\"")
        questions_continue = [
            inquirer.Confirm("continue", message="Do you confirm?", default=True)
        ]

        answers = inquirer.prompt(questions_continue)
        if answers['continue'] == False:
            print("Aborted!")
            exit(-1)

        logger.info(f"Starting label modification...")

        # add new label into the available-labels list (if not empty)
        self.add_label_to_available_list(label_to_add=label_new_value)

        self.apply_new_label(new_label=label_new_value, old_label=label_to_modify)
        logger.info(f"Applied new label.")

        self.modify_permissions(new_label=label_new_value, old_label=label_to_modify)
        logger.info(f"Modified permissions.")

        # remove old label from the available-labels list
        self.remove_label_from_available_list(label_to_remove=label_to_modify)

        logger.info("End of label modification!")


if __name__ == '__main__':
    level = logging.INFO

    logging.basicConfig(level=level, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    parser = argparse.ArgumentParser(description="Ask user to choose a label to modify.")
    parser.add_argument('--orthanc_url', help="The url of the Orthanc to modifiy the label to.", default="http://localhost:8042")
    parser.add_argument('--orthanc_user', help="The user of the Orthanc to modifiy the label to.", default="demo")
    parser.add_argument('--orthanc_password', help="The password of the Orthanc to modifiy the label to.", default=None)

    parser.add_argument('--auth_url', help="The root url of the auth service to modifiy the permissions to.", default="http://localhost/auth-service/")
    parser.add_argument('--auth_user', help="The user of the auth service to modifiy the permissions to.", default="demo")
    parser.add_argument('--auth_password', help="The password of the auth service to modifiy the permissions to.", default=None)

    args = parser.parse_args()

    orthanc_url = args.orthanc_url
    orthanc_user = args.orthanc_user
    orthanc_password = args.orthanc_password

    auth_url = args.auth_url
    auth_user = args.auth_user
    auth_password = args.auth_password

    # password should always be avoided as an argument of the command, so that it is not in the terminal history
    if orthanc_password is None:
        questions = [
            inquirer.Text('orthanc_password', message="Please enter Orthanc password")
        ]
        answers = inquirer.prompt(questions)
        orthanc_password = answers["orthanc_password"]

    if auth_password is None:
        questions = [
            inquirer.Text('auth_password', message="Please enter Auth-Service password")
        ]
        answers = inquirer.prompt(questions)
        auth_password = answers["auth_password"]

    orthanc_client=OrthancApiClient(orthanc_url, user=orthanc_user, pwd=orthanc_password)

    if auth_url[-1] == "/":
        auth_url = auth_url[:-1]
    modifier = LabelModifier(api_client=orthanc_client, auth_service_url=auth_url, auth_service_login=auth_user, auth_service_password=auth_password)

    # will crash if password is wrong
    modifier.get_roles()

    modifier.execute()



import time, sys
import argparse
import logging
from orthanc_api_client import OrthancApiClient
import os

# examples:
# python orthanc_tools/orthanc_downloader.py --folder=./backup --url=http://192.168.0.10:8042 --user=user --password=pwd --labels=label1,label2


logger = logging.getLogger(__name__)

if __name__ == '__main__':
    level = logging.INFO
    if os.environ.get('VERBOSE_ENABLED'):
        level = logging.DEBUG
    logging.basicConfig(level=level, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    parser = argparse.ArgumentParser(description='Download the content of Orthanc in a folder, according to the labels')
    parser.add_argument('--url', type=str, default='http://localhost:8042', help='Orthanc url')
    parser.add_argument('--user', type=str, default=None, help='Orthanc user name')
    parser.add_argument('--password', type=str, default=None, help='Orthanc password')
    parser.add_argument('--api_key', type=str, default=None, help='Orthanc api-key')
    parser.add_argument('--folder', type=str, help='Folder to store the DICOM files to.')
    parser.add_argument('--labels', type=str, default=None, help='Comma separated list of labels. Only studies with these labels will be downloaded.')
    args = parser.parse_args()

    o = None
    if args.api_key is not None:
        o=OrthancApiClient(args.url, headers={"api-key":args.api_key})
    else:
        o=OrthancApiClient(args.url, user=args.user, pwd=args.password)

    labels = []
    # labels args present, download only these ones
    if args.labels is not None:
        labels = args.labels.split(',')

    retry_count = 0
    retry_delays = [5, 20, 60, 300, 900]

    studies_ids = []
    while retry_count < 5:
        if retry_count >= 1:
            delay = retry_delays[retry_count - 1]
            logger.info(f"waiting {delay} seconds before retrying get studies ids")
            time.sleep(delay)
        try:
            logger.info(f"Getting list of studies ids...")
            # workaround to avoid find results limit
            all_ids = o.studies.get_all_ids()
            for id in all_ids:
                study_labels = o.studies.get_labels(id)
                if any(element in study_labels for element in labels):
                    studies_ids.append(id)
            break
        except Exception as ex:
            retry_count += 1
            if retry_count == 5:
                logger.error(f"Error (retried 5 times) while getting studies ids. Ex: {str(ex)}")
                sys.exit(1)
            else:
                logger.warning(f"Error while getting studies ids, retrying... Ex:{str(ex)}")

    progress_counter = 1
    for id in studies_ids:
        try:
            logger.info(f"Downloading study {progress_counter} out of {len(studies_ids)}")
            folder_path = args.folder + id
            os.mkdir(folder_path)
            o.studies.download_instances(study_id=id, path=folder_path)
            progress_counter += 1
        except Exception as ex:
            logger.error(f"Error during download for study {id}. Ex: {str(ex)}")
            sys.exit(1)

    logger.info("Over :-)")